package moa.classifiers;

import java.util.ArrayList;
import java.util.List;

import moa.core.DoubleVector;
import moa.core.Measurement;
import moa.options.ClassOption;
import moa.options.FloatOption;
import moa.options.IntOption;
import weka.core.Instance;
import weka.core.Utils;

/**
 * Dynamic weighted majority algorithm.
 * 
 * Extends the Weighted Majority Algorithm to add and remove experts based on
 * local and global accuracy.
 * 
 * <p>J. Zico Kolter and Marcus A. Maloof. Dynamic weighted majority: An ensemble method 
 * for drifting concepts. The Journal of Machine Learning Research, 8:2755-2790, December 
 * 2007. ISSN 1532-4435. URL http://dl.acm.org/citation.cfm?id=1314498.1390333.</p>
 * 
 * @author Paulo Gonalves (paulomgj@gmail.com)
 *
 */
public class DynamicWeightedMajority extends AbstractClassifier {

	public ClassOption baseLearnerOption = new ClassOption("baseLearner", 'l',
            "Classifier to train.", Classifier.class, "HoeffdingTreeNB");

	public IntOption periodOption = new IntOption("period", 'p',
            "Period between expert removal, creation, and weight update.", 50, 1, Integer.MAX_VALUE);

    public FloatOption betaOption = new FloatOption("beta", 'b',
            "Factor to punish mistakes by.", 0.5, 0.0, 1.0);

    public FloatOption gammaOption = new FloatOption("gamma", 'g',
            "Minimum fraction of weight per model.", 0.01, 0.0, 0.5);

    protected List<Classifier> ensemble;
    protected List<Double> ensembleWeights;
    protected long index;

    @Override
    public void resetLearningImpl() {
    	this.ensemble = new ArrayList<Classifier>();
    	Classifier classifier = (Classifier) getPreparedClassOption(this.baseLearnerOption);
    	classifier.resetLearning();
        this.ensemble.add(classifier);
    	this.ensembleWeights = new ArrayList<Double>();
    	this.ensembleWeights.add(1.0);
    	this.index = 0;
    }
    
	@Override
    public void trainOnInstanceImpl(Instance inst) {
		this.index++;
        double totalWeight = 0.0;
        double[] localPrediction = new double[inst.numClasses()];
        // Loop over experts
        for (int j = 0; j < this.ensemble.size(); j++) {
        	double[] sigma = this.ensemble.get(j).getVotesForInstance(inst);
        	int classifierVote = Utils.maxIndex(sigma);
        	if((classifierVote != (int) inst.classValue()) && this.index % this.periodOption.getValue() == 0) {
                this.ensembleWeights.set(j, this.ensembleWeights.get(j) * this.betaOption.getValue());	
        	}
        	localPrediction[classifierVote] += this.ensembleWeights.get(j);
            totalWeight += this.ensembleWeights.get(j);
        }
        int globalPrediction = Utils.maxIndex(localPrediction);
        if(this.index % this.periodOption.getValue() == 0) {
        	this.index = 0;
            // normalize weights
            for (int i = 0; i < this.ensembleWeights.size(); i++) {
                this.ensembleWeights.set(i, this.ensembleWeights.get(i) / totalWeight);
            }
            // remove experts except for the last one
            for (int i = 0; i < this.ensembleWeights.size() - 1; i++) {
            	if(this.ensembleWeights.get(i) < this.gammaOption.getValue()) {
            		this.ensemble.remove(i);
            		this.ensembleWeights.remove(i);
            	}
            }
            if(globalPrediction != (int) inst.classValue()) {
            	Classifier classifier = (Classifier) getPreparedClassOption(this.baseLearnerOption);
            	classifier.resetLearning();
                this.ensemble.add(classifier);
            	this.ensembleWeights.add(1.0);
            }
        }
        // train experts
        for (int i = 0; i < this.ensemble.size(); i++) {
        	this.ensemble.get(i).trainOnInstance(inst);
        }
    }

	@Override
	public boolean isRandomizable() {
		return false;
	}

	@Override
    public double[] getVotesForInstance(Instance inst) {
        DoubleVector combinedVote = new DoubleVector();
        if (this.trainingWeightSeenByModel > 0.0) {
            for (int i = 0; i < this.ensemble.size(); i++) {
                if (this.ensembleWeights.get(i) > 0.0) {
                    DoubleVector vote = new DoubleVector(this.ensemble.get(i).getVotesForInstance(inst));
                    if (vote.sumOfValues() > 0.0) {
                        vote.normalize();
                        vote.scaleValues(this.ensembleWeights.get(i));
                        combinedVote.addValues(vote);
                    }
                }
            }
        }
        return combinedVote.getArrayRef();
    }

    @Override
    protected Measurement[] getModelMeasurementsImpl() {
        Measurement[] measurements = null;
        if (this.ensembleWeights != null) {
            measurements = new Measurement[this.ensembleWeights.size()];
            for (int i = 0; i < this.ensembleWeights.size(); i++) {
                measurements[i] = new Measurement("member weight " + (i + 1),
                        this.ensembleWeights.get(i));
            }
        }
        return measurements;
    }

	@Override
	public void getModelDescription(StringBuilder out, int indent) {
		
	}
}
