package moa.classifiers;

import java.util.ArrayList;
import java.util.List;

import moa.core.DoubleVector;
import moa.core.Measurement;
import moa.options.ClassOption;
import moa.options.IntOption;
import weka.core.Instance;
import weka.core.Instances;

/**
 * Learning in non-stationary environments.
 * 
 * <p>
 * Michael Muhlbaier and Robi Polikar. An ensemble approach for incremental
 * learning in nonstationary environments. In Michal Haindl, Josef Kittler, and
 * Fabio Roli, editors, Multiple Classifier Systems, volume 4472 of Lecture
 * Notes in Computer Science, pages 490-500. Springer Berlin / Heidelberg, 2007.
 * ISBN 978-3-540-72481-0. URL http://dx.doi.org/10.1007/978-3-540-72523-7_49.
 * 10.1007/978-3-540-72523-7_49
 * </p>
 * 
 * @author Paulo Gonalves (paulomgj@gmail.com)
 * 
 */
public class LearnNSE extends AbstractClassifier {

	public ClassOption baseLearnerOption = new ClassOption("baseLearner", 'l',
			"Classifier to train.", Classifier.class, "NaiveBayes");

	public IntOption periodOption = new IntOption("period", 'p',
			"Size of the environments.", 250, 1, Integer.MAX_VALUE);

	protected List<Classifier> ensemble;
	protected List<Double> ensembleWeights;
	protected List<ArrayList<Double>> bkts;
	protected List<ArrayList<Double>> wkts;
	protected Instances buffer;
	protected long index;

	@Override
	public void resetLearningImpl() {
		this.ensemble = new ArrayList<Classifier>();
		this.ensembleWeights = new ArrayList<Double>();
		this.bkts = new ArrayList<ArrayList<Double>>();
		this.wkts = new ArrayList<ArrayList<Double>>();
		this.index = 0;
		this.buffer = null;
	}

	@Override
	public void trainOnInstanceImpl(Instance inst) {
		this.index++;
		// Store instance in the buffer
		if(this.buffer == null) this.buffer = new Instances(inst.dataset());
		this.buffer.add(inst);
		if (this.index % this.periodOption.getValue() == 0) {
			this.index = 0;
			double mt = this.buffer.numInstances();
			Classifier classifier = (Classifier) getPreparedClassOption(this.baseLearnerOption);
			classifier.resetLearning();
			if(this.ensemble.size() > 1) {
				double et = 0;
				// Reading all data chunk instances
				for (int i = 0; i < mt; i++) {
					// Compute current ensemble error on new data
					boolean vote = this.correctlyClassifies(this.buffer.instance(i));
					if(!vote) et += 1/mt;
				}
				System.out.println("et = " + et);
				// Normalizing error
				double weightSum = 0;
				// Reading all data chunk instances
				for (int i = 0; i < mt; i++) {
					Instance instance = this.buffer.instance(i);
					// Updating instance weights
					boolean vote = this.correctlyClassifies(instance);
					double error = (1/mt)*(vote ? et : 1);
					instance.setWeight(error);
					weightSum += error;
				}
				System.out.println("weightSum = " + weightSum);
				// Reading all data chunk instances
				for (int i = 0; i < mt; i++) {
					Instance instance = this.buffer.instance(i);
					// Normalize weights
					instance.setWeight(instance.weight()/weightSum);
					// Call base classifier
					classifier.trainOnInstance(instance);
				}
			} else {
				// First run! Iterating through all instances in the data chunk
				for (int i = 0; i < mt; i++) {
					Instance instance = this.buffer.instance(i);
					// Initialize weights
					instance.setWeight(1/mt);
					// Call base classifier
					classifier.trainOnInstance(instance);
				}
			}
			this.ensemble.add(classifier);
			this.bkts.add(new ArrayList<Double>());
			this.wkts.add(new ArrayList<Double>());
			this.ensembleWeights.clear();
			int t = this.ensemble.size();
			// Evaluate all existing classifiers on new data set
			for (int k = 0; k < t; k++) {
				System.out.println("k = " + k);
				double ekt = 0;
				// Reading all data chunk instances
				for (int i = 0; i < mt; i++) {
					Instance instance = this.buffer.instance(i);
					if (!this.ensemble.get(k).correctlyClassifies(instance)) {
						// Ensemble incorrectly classifies this instance
						ekt += instance.weight();
					}
				}
				if (k == t - 1 && ekt > 0.5) {
					// Generate a new classifier
					Classifier c = (Classifier) getPreparedClassOption(this.baseLearnerOption);
					c.resetLearning();
					this.ensemble.set(k, c);
				} else if (ekt > 0.5) {
					// Remove voting power of this classifier
					ekt = 0.5;
				}
				System.out.println("ekt = " + ekt);
				// Normalizing errors
				double bkt = ekt/(1-ekt);
				// Retrieving normalized errors for this classifier
				ArrayList<Double> nbkt = this.bkts.get(k);
				nbkt.add(bkt);
				System.out.println("bkt = " + bkt);
				// Compute a weighted sum of all normalized errors for kth classifier h_k
				double wkt = 1/(1+Math.exp(-0.05*(t-k-1)));
				List<Double> weights = this.wkts.get(k);
				//weights.add(wkt);
				double sum = 0;
				for(int j = 0; j < weights.size(); j++) {
					sum += weights.get(j);
				}
				//weights.remove(k);
				weights.add(wkt/(sum+wkt));
				System.out.println("sum = " + sum);
				System.out.println("wkt = " + wkt);
				double sbkt = 0;
				for(int j = 0; j < weights.size(); j++) {
					sbkt += weights.get(j)*nbkt.get(j);
				}
				// Calculate classifier voting weights
				this.ensembleWeights.add(Math.log10(1/sbkt));
				System.out.println("ensembleWeight = " + Math.abs(Math.log10(1/sbkt)));
			}
			this.buffer = new Instances(this.getModelContext());
		}
	}

	@Override
	public boolean isRandomizable() {
		return false;
	}

	@Override
    public double[] getVotesForInstance(Instance inst) {
        DoubleVector combinedVote = new DoubleVector();
        if (this.trainingWeightSeenByModel > 0.0) {
            for (int i = 0; i < this.ensemble.size(); i++) {
                if (this.ensembleWeights.get(i) > 0.0) {
                    DoubleVector vote = new DoubleVector(this.ensemble.get(i).getVotesForInstance(inst));
                    if (vote.sumOfValues() > 0.0) {
                        vote.normalize();
                        vote.scaleValues(this.ensembleWeights.get(i));
                        combinedVote.addValues(vote);
                    }
                }
            }
        }
        return combinedVote.getArrayRef();
    }

	@Override
	public void getModelDescription(StringBuilder out, int indent) {
	}

	@Override
	protected Measurement[] getModelMeasurementsImpl() {
		return null;
	}
}
