/*
 *    StaticWindow.java
 *    Copyright (C) 2011 Poznan University of Technology, Poland
 *    @author Dariusz Brzezinski (dariusz.brzezinski at cs dot put dot poznan dot pl)
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package moa.classifiers;

import moa.core.Measurement;
import moa.options.ClassOption;
import moa.options.IntOption;
import weka.core.Instance;

/**
 * A class for creating a static window out of a MOA classifier.
 * 
 * @author Dariusz Brzeziski
 */
public class StaticWindow extends AbstractClassifier
{
	/**
	 * Option for defining the window's base classifier.
	 */
	public ClassOption baseLearnerOption = new ClassOption("baseLearner", 'l', "Classifier to train.", Classifier.class, "trees.HoeffdingTree -e 2000000 -g 100 -c 0.01");

	/**
	 * Option for defining the width of the window.
	 */
	public IntOption widthOption = new IntOption("width", 'w', "Size of Window for training learner.", 0, 0, Integer.MAX_VALUE);

	/**
	 * Trained classifier.
	 */
	protected Classifier classifier, candidateClassifier;

	/**
	 * Number of processed examples/instances.
	 */
	protected int processedInstances;

	/**
	 * Determines whether the first classifier has already been trained. This variable should state false only for the first window of examples.
	 */
	protected boolean isClassificationEnabled;
	
	private static final long serialVersionUID = 1L;

	@Override
	public int measureByteSize()
	{
		return this.classifier.measureByteSize() + this.candidateClassifier.measureByteSize();
	}

	@Override
	public void resetLearningImpl()
	{
		try
		{
			this.classifier = (Classifier) getPreparedClassOption(this.baseLearnerOption);
			this.classifier.resetLearning();
			this.candidateClassifier = (Classifier) getPreparedClassOption(this.baseLearnerOption);
			this.candidateClassifier.resetLearning();
		}
		catch (Exception e)
		{
			System.err.println("Creating a new classifier: " + e.getMessage());
		}
		processedInstances = 0;
		isClassificationEnabled = false;
	}

	@Override
	public void trainOnInstanceImpl(Instance inst)
	{
		this.candidateClassifier.trainOnInstance(inst);
		this.processedInstances++;

		//// Create a new classifier after a window of examples
		if (this.processedInstances % this.widthOption.getValue() == 0)
		{
			this.classifier = this.candidateClassifier.copy();
			
			this.candidateClassifier = (Classifier) getPreparedClassOption(this.baseLearnerOption);
			this.candidateClassifier.resetLearning();
			
			//// Enable testing after the first window of examples
			if(!isClassificationEnabled){
				isClassificationEnabled = true;
			}
		}
	}

	public double[] getVotesForInstance(Instance inst)
	{
		double[] votes = new double[inst.numClasses()];
		if (isClassificationEnabled == false)
		{
			for (int i = 0; i < inst.numClasses(); i++)
			{
				votes[i] = 1.0 / inst.numClasses();
			}
		}
		else
		{
			try
			{
				votes = this.classifier.getVotesForInstance(inst);
			}
			catch (Exception e)
			{
				System.err.println(e.getMessage());
			}
		}
		return votes;
	}

	/**
	 * The static window is not randomizable.
	 */
	public boolean isRandomizable()
	{
		return false;
	}

	@Override
	public void getModelDescription(StringBuilder out, int indent)
	{
		if (classifier != null)
		{
			out.append(classifier.toString());
		}
	}

	@Override
	protected Measurement[] getModelMeasurementsImpl()
	{
		Measurement[] m = new Measurement[0];
		return m;
	}
}
