/*
 *    AccuracyUpdatedEnsemble.java
 *    Copyright (C) 2010 Poznan University of Technology, Poznan, Poland
 *    @author Dariusz Brzezinski (dariusz.brzezinski@cs.put.poznan.pl)
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program. If not, see <http://www.gnu.org/licenses/>.
 *    
 */
package moa.classifiers.meta;

import moa.classifiers.Classifier;
import moa.classifiers.core.driftdetection.DriftDetectionMethod;
import moa.options.ClassOption;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.Utils;

/**
 * The online version of the drift detector modification of the Accuracy Updated Ensemble 
 * as proposed by Brzezinski and Stefanowski in "From Block-based Ensembles 
 * to Online Learners In Changing Data Streams: If- and How-To", IID, ECML PKDD 2021.
 */
public class IncrementalAUE1DetectorOnline extends IncrementalAWEDetector {

    private static final long serialVersionUID = 1L;

    public IncrementalAUE1DetectorOnline()
    {
    	this.learnerOption = new ClassOption("learner", 'l', "Classifier to train.", Classifier.class, "trees.HoeffdingTree -e 1000 -g 100 -c 0.01");
    }
    
    @Override
    public String getPurposeString() {
        return "Accuracy Updated Ensemble classifier as proposed by Brzezinski and Stefanowski in 'Accuracy Updated Ensemble for Data Streams with Concept Drift'";
    }

    @Override
	public void trainOnInstanceImpl(Instance inst) {
		this.initVariables();

		this.classDistributions[(int) inst.classValue()]++;
		this.currentChunk.add(inst);
		this.detectorClassifier.trainOnInstance(inst);
		
		double mse_r = this.computeMseR();
    	
    	for (int i = 0; i < this.ensemble.length; i++) {
            if (mse_r > 0 && this.ensembleWeights[i] > 1.0 / mse_r) {
				this.ensemble[i].trainOnInstance(inst);
			}
        }
		
		this.processedInstances++;

		int trueClass = (int) inst.classValue();
		boolean prediction;
		if (Utils.maxIndex(this.detectorClassifier.getVotesForInstance(inst)) == trueClass) {
			prediction = true;
		} else {
			prediction = false;
		}
		this.ddmLevel = this.driftDetectionMethod.computeNextVal(prediction);

		switch (this.ddmLevel) {
		case DriftDetectionMethod.DDM_WARNING_LEVEL:
			if (newClassifierReset == true) {
				this.alarmDetectorClassifier.resetLearning();
				newClassifierReset = false;
			}
			this.alarmDetectorClassifier.trainOnInstance(inst);
			break;

		case DriftDetectionMethod.DDM_OUTCONTROL_LEVEL:
			this.isAlarm = true;
			this.detectorClassifier = null;
			this.detectorClassifier = this.alarmDetectorClassifier;
			if (this.detectorClassifier instanceof WEKAClassifier) {
				((WEKAClassifier) this.detectorClassifier).buildClassifier();
			}

			this.detectorWeight = this.computeWeight(this.detectorClassifier, this.currentChunk);
					
			this.processChunk();		
			
			// reset detector classifiers
			this.alarmDetectorClassifier = ((Classifier) getPreparedClassOption(this.detectorLearnerOption)).copy();
			this.alarmDetectorClassifier.resetLearning();

			// reset chunk
    		this.classDistributions = null;
    		this.currentChunk = null;
    		this.processedInstances = 0;
    		this.initVariables();
    		this.classDistributions[(int) inst.classValue()]++;
    		this.currentChunk.add(inst);
    		this.detectorClassifier.trainOnInstance(inst);
    		this.processedInstances++;
			this.isAlarm = false;
			break;

		case DriftDetectionMethod.DDM_INCONTROL_LEVEL:
			newClassifierReset = true;
			break;
		}

		if (this.processedInstances % this.chunkSize == 0) {
			this.processChunk();
		}
	}
    
    
    @Override
    protected void processChunk() {
        Classifier addedClassifier = null;
        double candidateClassifierWeight;
		
        if(this.currentChunk.numInstances() > this.numFolds){
			candidateClassifierWeight= this.computeCandidateWeight(this.candidateClassifier, this.currentChunk, this.numFolds);
		} else{
			candidateClassifierWeight = 1;
		}

        for (int i = 0; i < this.storedLearners.length; i++) {
            this.storedWeights[i][0] = this.computeWeight(
                    this.storedLearners[(int) this.storedWeights[i][1]], this.currentChunk);
        }

        if (this.storedLearners.length < this.maxStoredCount) {
            // Train and add classifier
            this.trainOnChunk(this.candidateClassifier, this.currentChunk);
            addedClassifier = this.addToStored(this.candidateClassifier,
                    candidateClassifierWeight);
        } else {
            // Substitute poorest classifier
            java.util.Arrays.sort(this.storedWeights, weightComparator);

            if (this.storedWeights[0][0] < candidateClassifierWeight) {
                this.trainOnChunk(this.candidateClassifier, this.currentChunk);
                this.storedWeights[0][0] = candidateClassifierWeight;
                addedClassifier = this.candidateClassifier.copy();
                this.storedLearners[(int) this.storedWeights[0][1]] = addedClassifier;
            }
        }

        int ensembleSize = java.lang.Math.min(this.storedLearners.length,
                this.maxMemberCount);
        this.ensemble = new Classifier[ensembleSize];
        this.ensembleWeights = new double[ensembleSize];

        // Sort learners according to their weights
        java.util.Arrays.sort(this.storedWeights, weightComparator);

        // Select top k classifiers to construct the ensemble
        int storeSize = this.storedLearners.length;
        for (int i = 0; i < ensembleSize; i++) {
            this.ensembleWeights[i] = this.storedWeights[storeSize - i - 1][0];
            this.ensemble[i] = this.storedLearners[(int) this.storedWeights[storeSize
                    - i - 1][1]];
        }

        this.classDistributions = null;
        this.currentChunk = null;
        this.candidateClassifier = (Classifier) getPreparedClassOption(this.learnerOption);
        this.candidateClassifier.resetLearning();
    }
    
    @Override
    protected double computeWeight(Classifier learner, Instances chunk) {
        double mse_i = 0;

        double f_ci;
        double voteSum;

        for (int i = 0; i < chunk.numInstances(); i++) {
            try {
                voteSum = 0;
                for (double element : learner.getVotesForInstance(chunk.instance(i))) {
                    voteSum += element;
                }

                if (voteSum > 0) {
                    f_ci = learner.getVotesForInstance(chunk.instance(i))[(int) chunk.instance(i).classValue()] / voteSum;
                    mse_i += (1 - f_ci) * (1 - f_ci);
                } else {
                    mse_i += 1;
                }
            } catch (Exception e) {
                mse_i += 1;
            }
        }

        mse_i /= chunk.numInstances();

        if (mse_i > 0) {
            return 1.0 / mse_i;
        } else {
            return Double.MAX_VALUE;
        }
    }

    /**
     * Trains a component classifier on the most recent chunk of data.
     *
     * @param classifierToTrain Classifier being trained.
     */
    private void trainOnChunk(Classifier classifierToTrain, Instances chunk) {
        for (int num = 0; num < chunk.numInstances(); num++) {
            classifierToTrain.trainOnInstance(chunk.instance(num));
        }
    }

    /**
     * Determines whether the classifier is randomizable.
     */
    public boolean isRandomizable() {
        return false;
    }
}
