/*
 *    AccuracyWeightedEnsemble.java
 *    Copyright (C) 2010 Poznan University of Technology, Poznan, Poland
 *    @author Dariusz Brzezinski (dariusz.brzezinski@cs.put.poznan.pl)
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program. If not, see <http://www.gnu.org/licenses/>.
 *    
 */
package moa.classifiers.meta;

import moa.classifiers.Classifier;
import moa.core.DoubleVector;
import moa.core.ObjectRepository;
import moa.options.ClassOption;
import moa.tasks.TaskMonitor;
import weka.core.Instance;
import weka.core.Instances;

/**
 * The candidate classifier modification of the Accuracy Weighted Ensemble 
 * as proposed by Brzezinski and Stefanowski in "From Block-based Ensembles 
 * to Online Learners In Changing Data Streams: If- and How-To", IID, ECML PKDD 2021.
 */
public class IncrementalAWECandidate extends AccuracyWeightedEnsemble {

	private static final long serialVersionUID = 1L;

	/**
	 * Type of classifier to use as a component classifier.
	 */
	public ClassOption candidateOption = new ClassOption("candidate", 'k',
			"Candidate classifier.", Classifier.class,
			"trees.HoeffdingTree -e 1000 -g 100 -c 0.01");

	@Override
	public void prepareForUseImpl(TaskMonitor monitor,
			ObjectRepository repository) {
		super.prepareForUseImpl(monitor, repository);

		this.candidateClassifier = (Classifier) getPreparedClassOption(this.candidateOption);
		this.candidateClassifier.resetLearning();
	}

	@Override
	public void resetLearningImpl() {
		super.resetLearningImpl();

		this.candidateClassifier = (Classifier) getPreparedClassOption(this.candidateOption);
		this.candidateClassifier.resetLearning();
	}

	@Override
	public void trainOnInstanceImpl(Instance inst) {
		this.initVariables();

		this.classDistributions[(int) inst.classValue()]++;
		this.currentChunk.add(inst);
		this.processedInstances++;

		// Incrementally train the candidate classifier
		this.candidateClassifier.trainOnInstance(inst);

		if (this.processedInstances % this.chunkSize == 0) {
			this.processChunk();
		}
	}

	/**
	 * Initiates the current chunk and class distribution variables.
	 */
	protected void initVariables() {
		if (this.currentChunk == null) {
			this.currentChunk = new Instances(this.getModelContext());
		}

		if (this.classDistributions == null) {
			this.classDistributions = new long[this.getModelContext()
					.classAttribute().numValues()];

			for (int i = 0; i < this.classDistributions.length; i++) {
				this.classDistributions[i] = 0;
			}
		}
	}

	/**
	 * Processes a chunk.
	 * 
	 * @param useMseR
	 *            Determines whether to use the MSEr threshold.
	 */
	protected void processChunk() {

		Classifier newClassifier = (Classifier) getPreparedClassOption(this.learnerOption);
		newClassifier.resetLearning();

		// Compute weights
		double newClassifierWeight = this.computeCandidateWeight(newClassifier,
				this.currentChunk, this.numFolds);

		for (int i = 0; i < this.storedLearners.length; i++) {
			this.storedWeights[i][0] = this.computeWeight(
					this.storedLearners[(int) this.storedWeights[i][1]],
					this.currentChunk);
		}

		if (this.storedLearners.length < this.maxStoredCount) {
			// Train and add classifier
			for (int num = 0; num < this.chunkSize; num++) {
				newClassifier.trainOnInstance(this.currentChunk.instance(num));
			}

			this.addToStored(newClassifier, newClassifierWeight);
		} else {
			// Substitute poorest classifier
			java.util.Arrays.sort(this.storedWeights, weightComparator);

			if (this.storedWeights[0][0] < newClassifierWeight) {
				for (int num = 0; num < this.chunkSize; num++) {
					newClassifier.trainOnInstance(this.currentChunk
							.instance(num));
				}

				this.storedWeights[0][0] = newClassifierWeight;
				this.storedLearners[(int) this.storedWeights[0][1]] = newClassifier
						.copy();
			}
		}

		int ensembleSize = java.lang.Math.min(this.storedLearners.length,
				this.maxMemberCount);
		this.ensemble = new Classifier[ensembleSize];
		this.ensembleWeights = new double[ensembleSize];

		// Sort learners according to their weights
		java.util.Arrays.sort(this.storedWeights, weightComparator);

		// Select top k classifiers to construct the ensemble
		int storeSize = this.storedLearners.length;
		for (int i = 0; i < ensembleSize; i++) {
			this.ensembleWeights[i] = this.storedWeights[storeSize - i - 1][0];
			this.ensemble[i] = this.storedLearners[(int) this.storedWeights[storeSize
					- i - 1][1]];
		}

		// Reset candidate classifier
		this.candidateClassifier = (Classifier) getPreparedClassOption(this.candidateOption);
		this.candidateClassifier.resetLearning();

		// Initialize the candidate with examples from the last chunk
		for (int num = 0; num < this.chunkSize; num++) {
			this.candidateClassifier.trainOnInstance(this.currentChunk
					.instance(num));
		}

		this.classDistributions = null;
		this.currentChunk = null;
	}

	/**
	 * Predicts a class for an example.
	 */
	@Override
	public double[] getVotesForInstance(Instance inst) {
		DoubleVector combinedVote = new DoubleVector();

		if (this.trainingWeightSeenByModel > 0.0) {
			for (int i = 0; i < this.ensemble.length; i++) {
				if (this.ensembleWeights[i] > 0.0) {
					DoubleVector vote = new DoubleVector(
							this.ensemble[i].getVotesForInstance(inst));

					if (vote.sumOfValues() > 0.0) {
						vote.normalize();
						// scale weight and prevent overflow
						vote.scaleValues(this.ensembleWeights[i]
								/ (1.0 * this.ensemble.length + 2));
						combinedVote.addValues(vote);
					}
				}
			}

			// Take the candidate classifiers vote into account
			DoubleVector cadidatevote = new DoubleVector(
					this.candidateClassifier.getVotesForInstance(inst));

			if (cadidatevote.sumOfValues() > 0.0) {
				cadidatevote.normalize();
				// scale weight and prevent overflow
				cadidatevote.scaleValues(this.staticCandidateWeight()
						/ (1.0 * this.ensemble.length + 2));
				combinedVote.addValues(cadidatevote);
			}
		}
		combinedVote.normalize();
		return combinedVote.getArrayRef();
	}

	private double candidateWeight = 0;

	protected double staticCandidateWeight() {
		if (this.classDistributions != null) {
			return recalculateIncrementalCandidateWeight();
		} else {
			return this.candidateWeight;
		}
	}

	protected double recalculateIncrementalCandidateWeight() {
		if (this.ensembleWeights != null && this.ensembleWeights.length > 0) {
			// Get the maximum from previous ensemble components
			double maximum = this.ensembleWeights[0];
			for (int i = 1; i < this.ensembleWeights.length; i++) {
				if (this.ensembleWeights[i] > maximum) {
					maximum = this.ensembleWeights[i];
				}
			}
			this.candidateWeight = maximum;
			return this.candidateWeight;
		} else {
			return 1;
		}
	}
}
