/*
 *    AccuracyWeightedEnsemble.java
 *    Copyright (C) 2010 Poznan University of Technology, Poznan, Poland
 *    @author Dariusz Brzezinski (dariusz.brzezinski@cs.put.poznan.pl)
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program. If not, see <http://www.gnu.org/licenses/>.
 *    
 */
package moa.classifiers.meta;

import moa.classifiers.Classifier;
import moa.classifiers.core.driftdetection.DriftDetectionMethod;
import moa.core.DoubleVector;
import moa.core.ObjectRepository;
import moa.options.ClassOption;
import moa.tasks.TaskMonitor;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.Utils;

/**
 * The drift detector modification of the Accuracy Updated Ensemble 
 * as proposed by Brzezinski and Stefanowski in "From Block-based Ensembles 
 * to Online Learners In Changing Data Streams: If- and How-To", IID, ECML PKDD 2021.
 */
public class IncrementalAWEDetector extends AccuracyWeightedEnsemble {

	private static final long serialVersionUID = 1L;

	public ClassOption detectorLearnerOption = new ClassOption(
			"detectorBaseLearner", 'b', "Classifier to train.",
			Classifier.class, "bayes.NaiveBayes");

	public ClassOption driftDetectionMethodOption = new ClassOption(
			"driftDetectionMethod", 'd', "Drift detection method to use.",
			DriftDetectionMethod.class, "DDM");

	protected Classifier detectorClassifier;
	protected Classifier alarmDetectorClassifier;
	protected DriftDetectionMethod driftDetectionMethod;
	protected boolean newClassifierReset;
	protected boolean isAlarm;
	protected int ddmLevel;
	protected double detectorWeight;

	@Override
	public void prepareForUseImpl(TaskMonitor monitor,
			ObjectRepository repository) {
		super.prepareForUseImpl(monitor, repository);

		this.detectorClassifier = ((Classifier) getPreparedClassOption(this.detectorLearnerOption))
				.copy();
		this.detectorClassifier.resetLearning();
		this.alarmDetectorClassifier = ((Classifier) getPreparedClassOption(this.detectorLearnerOption))
				.copy();
		this.alarmDetectorClassifier.resetLearning();
		this.driftDetectionMethod = ((DriftDetectionMethod) getPreparedClassOption(this.driftDetectionMethodOption))
				.copy();
		this.newClassifierReset = false;
		this.isAlarm = false;
		this.detectorWeight = 1;
	}

	@Override
	public void resetLearningImpl() {
		super.resetLearningImpl();

		this.detectorClassifier = ((Classifier) getPreparedClassOption(this.detectorLearnerOption))
				.copy();
		this.detectorClassifier.resetLearning();
		this.alarmDetectorClassifier = ((Classifier) getPreparedClassOption(this.detectorLearnerOption))
				.copy();
		this.alarmDetectorClassifier.resetLearning();
		this.driftDetectionMethod = ((DriftDetectionMethod) getPreparedClassOption(this.driftDetectionMethodOption))
				.copy();
		this.newClassifierReset = false;
		this.isAlarm = false;
		this.detectorWeight = 1;
	}

	@Override
	public void trainOnInstanceImpl(Instance inst) {
		this.initVariables();

		this.classDistributions[(int) inst.classValue()]++;
		this.currentChunk.add(inst);
		this.detectorClassifier.trainOnInstance(inst);
		this.processedInstances++;

		int trueClass = (int) inst.classValue();
		boolean prediction;
		if (Utils.maxIndex(this.detectorClassifier.getVotesForInstance(inst)) == trueClass) {
			prediction = true;
		} else {
			prediction = false;
		}
		this.ddmLevel = this.driftDetectionMethod.computeNextVal(prediction);

		switch (this.ddmLevel) {
		case DriftDetectionMethod.DDM_WARNING_LEVEL:
			if (newClassifierReset == true) {
				this.alarmDetectorClassifier.resetLearning();
				newClassifierReset = false;
			}
			this.alarmDetectorClassifier.trainOnInstance(inst);
			break;

		case DriftDetectionMethod.DDM_OUTCONTROL_LEVEL:
			this.isAlarm = true;
			this.detectorClassifier = null;
			this.detectorClassifier = this.alarmDetectorClassifier;
			if (this.detectorClassifier instanceof WEKAClassifier) {
				((WEKAClassifier) this.detectorClassifier).buildClassifier();
			}

			this.detectorWeight = this.computeWeight(this.detectorClassifier, this.currentChunk);
					
			this.processChunk();		
			
			// reset detector classifiers
			this.alarmDetectorClassifier = ((Classifier) getPreparedClassOption(this.detectorLearnerOption)).copy();
			this.alarmDetectorClassifier.resetLearning();

			// reset chunk
    		this.classDistributions = null;
    		this.currentChunk = null;
    		this.processedInstances = 0;
    		this.initVariables();
    		this.classDistributions[(int) inst.classValue()]++;
    		this.currentChunk.add(inst);
    		this.detectorClassifier.trainOnInstance(inst);
    		this.processedInstances++;
			this.isAlarm = false;
			break;

		case DriftDetectionMethod.DDM_INCONTROL_LEVEL:
			newClassifierReset = true;
			break;
		}

		if (this.processedInstances % this.chunkSize == 0) {
			this.processChunk();
		}
	}

	/**
	 * Initiates the current chunk and class distribution variables.
	 */
	protected void initVariables() {
		if (this.currentChunk == null) {
			this.currentChunk = new Instances(this.getModelContext());
		}

		if (this.classDistributions == null) {
			this.classDistributions = new long[this.getModelContext()
					.classAttribute().numValues()];

			for (int i = 0; i < this.classDistributions.length; i++) {
				this.classDistributions[i] = 0;
			}
		}
	}

	protected void processChunk() {
        // Compute weights
		double candidateClassifierWeight;
		if(this.currentChunk.numInstances() > this.numFolds){
			candidateClassifierWeight= this.computeCandidateWeight(this.candidateClassifier, this.currentChunk, this.numFolds);
		} else{
			candidateClassifierWeight = 1;
		}

        for (int i = 0; i < this.storedLearners.length; i++) {
            this.storedWeights[i][0] = this.computeWeight(this.storedLearners[(int) this.storedWeights[i][1]], this.currentChunk);
        }

        if (this.storedLearners.length < this.maxStoredCount) {
            // Train and add classifier
            for (int num = 0; num < this.currentChunk.numInstances(); num++) {
                this.candidateClassifier.trainOnInstance(this.currentChunk.instance(num));
            }

            this.addToStored(this.candidateClassifier, candidateClassifierWeight);
        } else {
            // Substitute poorest classifier
            java.util.Arrays.sort(this.storedWeights, weightComparator);

            if (this.storedWeights[0][0] < candidateClassifierWeight) {
                for (int num = 0; num < this.currentChunk.numInstances(); num++) {
                    this.candidateClassifier.trainOnInstance(this.currentChunk.instance(num));
                }

                this.storedWeights[0][0] = candidateClassifierWeight;
                this.storedLearners[(int) this.storedWeights[0][1]] = this.candidateClassifier.copy();
            }
        }

        int ensembleSize = java.lang.Math.min(this.storedLearners.length, this.maxMemberCount);
        this.ensemble = new Classifier[ensembleSize];
        this.ensembleWeights = new double[ensembleSize];

        // Sort learners according to their weights
        java.util.Arrays.sort(this.storedWeights, weightComparator);

        // Select top k classifiers to construct the ensemble
        int storeSize = this.storedLearners.length;
        for (int i = 0; i < ensembleSize; i++) {
            this.ensembleWeights[i] = this.storedWeights[storeSize - i - 1][0];
            this.ensemble[i] = this.storedLearners[(int) this.storedWeights[storeSize - i - 1][1]];
        }

        this.classDistributions = null;
        this.currentChunk = null;
        this.candidateClassifier = (Classifier) getPreparedClassOption(this.learnerOption);
        this.candidateClassifier.resetLearning();
    }

	/**
	 * Computes the weight of a given classifier.
	 * 
	 * @param learner
	 *            Classifier to calculate weight for.
	 * @param chunk
	 *            Data chunk of examples.
	 * @param useMseR
	 *            Determines whether to use the MSEr threshold.
	 * @return The given classifier's weight.
	 */
	protected double computeWeight(Classifier learner, Instances chunk) {
		double mse_i = 0;
		double mse_r = 0;

		double f_ci;
		double voteSum;

		for (int i = 0; i < chunk.numInstances(); i++) {
			try {
				voteSum = 0;
				for (double element : learner.getVotesForInstance(chunk
						.instance(i))) {
					voteSum += element;
				}

				if (voteSum > 0) {
					f_ci = learner.getVotesForInstance(chunk.instance(i))[(int) chunk
							.instance(i).classValue()] / voteSum;
					mse_i += (1 - f_ci) * (1 - f_ci);
				} else {
					mse_i += 1;
				}
			} catch (Exception e) {
				mse_i += 1;
			}
		}

		mse_i /= chunk.numInstances();
		mse_r = this.computeMseR();

		return java.lang.Math.max(mse_r - mse_i, 0);
	}
	
	/**
	 * Predicts a class for an example.
	 */
	@Override
	public double[] getVotesForInstance(Instance inst) {
		DoubleVector combinedVote = new DoubleVector();

		if (this.trainingWeightSeenByModel > 0.0) {
			for (int i = 0; i < this.ensemble.length; i++) {
				if (this.ensembleWeights[i] > 0.0) {
					DoubleVector vote = new DoubleVector(
							this.ensemble[i].getVotesForInstance(inst));

					if (vote.sumOfValues() > 0.0) {
						vote.normalize();
						// scale weight and prevent overflow
						vote.scaleValues(this.ensembleWeights[i]
								/ (1.0 * this.ensemble.length + 2));
						combinedVote.addValues(vote);
					}
				}
			}

			// Take the candidate classifiers vote into account
			DoubleVector detectorVote = new DoubleVector(
					this.detectorClassifier.getVotesForInstance(inst));

			if (detectorVote.sumOfValues() > 0.0) {
				detectorVote.normalize();
				// scale weight and prevent overflow
				detectorVote.scaleValues(this.detectorWeight
						/ (1.0 * this.ensemble.length + 2));
				combinedVote.addValues(detectorVote);
			}
		}
		combinedVote.normalize();
		return combinedVote.getArrayRef();
	}
}
