/*
 *    AccuracyWeightedEnsemble.java
 *    Copyright (C) 2010 Poznan University of Technology, Poznan, Poland
 *    @author Dariusz Brzezinski (dariusz.brzezinski@cs.put.poznan.pl)
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program. If not, see <http://www.gnu.org/licenses/>.
 *    
 */
package moa.classifiers.meta;

import moa.classifiers.Classifier;
import moa.core.ObjectRepository;
import moa.options.IntOption;
import moa.tasks.TaskMonitor;
import weka.core.Instance;
import weka.core.Instances;

/**
 * The windowing modification of the Accuracy Weighted Ensemble 
 * as proposed by Brzezinski and Stefanowski in "From Block-based Ensembles 
 * to Online Learners In Changing Data Streams: If- and How-To", IID, ECML PKDD 2021.
 */
public class IncrementalAWEWindow extends AccuracyWeightedEnsemble {

	/**
	 * SUID.
	 */
	private static final long serialVersionUID = 8338538373947641905L;

	/**
	 * Constructor.
	 * Converts the chunkSize option into the windowSize option.
	 */
	public IncrementalAWEWindow()
	{
		this.chunkSizeOption = new IntOption("windowSize", 'c', "The window size used for candidate classifier creation.", 500, 1, Integer.MAX_VALUE);
	}

	
    @Override
    public String getPurposeString() {
        return "Accuracy Weighted Ensemble classifier as proposed by Wang et al. in 'Mining concept-drifting data streams using ensemble classifiers', KDD 2003";
    }

    protected int windowSize;

    protected Instances currentWindow;

    @Override
    public void prepareForUseImpl(TaskMonitor monitor, ObjectRepository repository) {
        this.maxMemberCount = (int) memberCountOption.getValue();
        this.maxStoredCount = (int) storedCountOption.getValue();

        if (this.maxMemberCount > this.maxStoredCount) {
            this.maxStoredCount = this.maxMemberCount;
        }

        this.windowSize = this.chunkSizeOption.getValue();
        this.numFolds = this.numFoldsOption.getValue();
        this.candidateClassifier = (Classifier) getPreparedClassOption(this.learnerOption);
        this.candidateClassifier.resetLearning();

        super.prepareForUseImpl(monitor, repository);
    }

    @Override
    public void resetLearningImpl() {
    	this.currentWindow = null;
        this.classDistributions = null;
        
        this.processedInstances = 0;
        this.ensemble = new Classifier[0];
        this.storedLearners = new Classifier[0];

        this.candidateClassifier = (Classifier) getPreparedClassOption(this.learnerOption);
        this.candidateClassifier.resetLearning();
    }

    @Override
    public void trainOnInstanceImpl(Instance inst) {
        this.initVariables();
    	
    	if(this.processedInstances < this.windowSize)
        {
    		this.classDistributions[(int) inst.classValue()]++;
    		this.currentWindow.add(inst);		
        }
    	else
        {
    		int windowPosition = processedInstances % this.windowSize;
    		
    		this.classDistributions[(int) this.currentWindow.instance(windowPosition).classValue()]++;
    		this.classDistributions[(int) inst.classValue()]++;
    		
        	this.currentWindow.set(windowPosition, inst);   	
        }
        
    	this.processedInstances++;
    	
        if (this.processedInstances % this.windowSize == 0) {
            this.createNewClassifier();
        } else {
        	for (int i = 0; i < this.ensemble.length; i++) {
                this.ensembleWeights[i] = this.computeWeight(this.ensemble[i], this.currentWindow);
            }
        }
    }

    /**
     * Initiates the current chunk and class distribution variables.
     */
    protected void initVariables() {
        if (this.currentWindow == null) {
            this.currentWindow = new Instances(this.getModelContext());
        }

        if (this.classDistributions == null) {
            this.classDistributions = new long[this.getModelContext().classAttribute().numValues()];

            for (int i = 0; i < this.classDistributions.length; i++) {
                this.classDistributions[i] = 0;
            }
        }
    }
    
    /**
     * Processes a chunk.
     *
     * @param useMseR Determines whether to use the MSEr threshold.
     */
    protected void createNewClassifier() {
        // Compute weights
        double candidateClassifierWeight = this.computeCandidateWeight(this.candidateClassifier, this.currentWindow, this.numFolds);

        for (int i = 0; i < this.storedLearners.length; i++) {
            this.storedWeights[i][0] = this.computeWeight(this.storedLearners[(int) this.storedWeights[i][1]], this.currentWindow);
        }

        if (this.storedLearners.length < this.maxStoredCount) {
            // Train and add classifier
            for (int num = 0; num < this.windowSize; num++) {
                this.candidateClassifier.trainOnInstance(this.currentWindow.instance(num));
            }

            this.addToStored(this.candidateClassifier, candidateClassifierWeight);
        } else {
            // Substitute poorest classifier
            java.util.Arrays.sort(this.storedWeights, weightComparator);

            if (this.storedWeights[0][0] < candidateClassifierWeight) {
                for (int num = 0; num < this.windowSize; num++) {
                    this.candidateClassifier.trainOnInstance(this.currentWindow.instance(num));
                }

                this.storedWeights[0][0] = candidateClassifierWeight;
                this.storedLearners[(int) this.storedWeights[0][1]] = this.candidateClassifier.copy();
            }
        }

        int ensembleSize = java.lang.Math.min(this.storedLearners.length, this.maxMemberCount);
        this.ensemble = new Classifier[ensembleSize];
        this.ensembleWeights = new double[ensembleSize];

        // Sort learners according to their weights
        java.util.Arrays.sort(this.storedWeights, weightComparator);

        // Select top k classifiers to construct the ensemble
        int storeSize = this.storedLearners.length;
        for (int i = 0; i < ensembleSize; i++) {
            this.ensembleWeights[i] = this.storedWeights[storeSize - i - 1][0];
            this.ensemble[i] = this.storedLearners[(int) this.storedWeights[storeSize - i - 1][1]];
        }

        this.candidateClassifier = (Classifier) getPreparedClassOption(this.learnerOption);
        this.candidateClassifier.resetLearning();
    }
}
