package moa.classifiers.meta;

import java.util.ArrayList;
import java.util.List;

import moa.classifiers.AbstractClassifier;
import moa.classifiers.Classifier;
import moa.core.DoubleVector;
import moa.core.Measurement;
import moa.options.ClassOption;
import moa.options.FloatOption;
import moa.options.IntOption;
import moa.options.MultiChoiceOption;
import weka.core.Instance;
import weka.core.Instances;

/**
 * Learning in non-stationary environments.
 * 
 * <p>
 * Ryan Elwell and Robi Polikar. Incremental learning of concept drift in
 * non-stationary environments. IEEE Transactions on Neural Networks,
 * 22(10):1517{1531, October 2011. ISSN 1045-9227. URL
 * http://dx.doi.org/10.1109/TNN.2011.2160459.
 * </p>
 * 
 * @author Paulo Gonalves (paulomgj@gmail.com)
 * 
 */
public class LearnNSE extends AbstractClassifier {

	public ClassOption baseLearnerOption = new ClassOption("baseLearner", 'l',
			"Classifier to train.", Classifier.class, "bayes.NaiveBayes");

	public IntOption periodOption = new IntOption("period", 'p',
			"Size of the environments.", 250, 1, Integer.MAX_VALUE);

	public FloatOption sigmoidSlopeOption = new FloatOption(
			"sigmoidSlope",
			'a',
			"Slope of the sigmoid function controlling the number "
					+ "of previous periods taken into account during weighting.",
			0.5, 0, Float.MAX_VALUE);

	public FloatOption sigmoidCrossingPointOption = new FloatOption(
			"sigmoidCrossingPoint",
			'b',
			"Halfway crossing point of the sigmoid function controlling the number of previous "
					+ "periods taken into account during weighting.", 10, 0,
			Float.MAX_VALUE);

	public IntOption ensembleSizeOption = new IntOption("ensembleSize", 'e',
			"Ensemble size.", 15, 1, Integer.MAX_VALUE);

	public MultiChoiceOption pruningStrategyOption = new MultiChoiceOption(
			"pruningStrategy", 's', "Classifiers pruning strategy to be used.",
			new String[] { "NO", "AGE", "ERROR" }, new String[] {
					"Don't prune classifiers", "Age-based", "Error-based" }, 0);

	protected List<Classifier> ensemble;
	protected List<Double> ensembleWeights;
	protected List<ArrayList<Double>> bkts, wkts;
	protected Instances buffer;
	protected long index;
	protected double slope, crossingPoint, maxError;
	protected int pruning, errorIndex, ensembleSize;

	@Override
	public void resetLearningImpl() {
		this.ensemble = new ArrayList<Classifier>();
		this.ensembleWeights = new ArrayList<Double>();
		this.bkts = new ArrayList<ArrayList<Double>>();
		this.wkts = new ArrayList<ArrayList<Double>>();
		this.index = 0;
		this.buffer = null;
		this.slope = sigmoidSlopeOption.getValue();
		this.crossingPoint = sigmoidCrossingPointOption.getValue();
		this.pruning = pruningStrategyOption.getChosenIndex();
		this.maxError = Double.MIN_VALUE;
		this.errorIndex = Integer.MIN_VALUE;
		this.ensembleSize = ensembleSizeOption.getValue();
	}

	@Override
	public void trainOnInstanceImpl(Instance inst) {
		this.index++;
		// Store instance in the buffer
		if (this.buffer == null) {
			this.buffer = new Instances(inst.dataset());
		}
		this.buffer.add(inst);

		if (this.index % this.periodOption.getValue() == 0) {
			this.index = 0;
			double mt = this.buffer.numInstances();
			Classifier classifier = ((Classifier) getPreparedClassOption(this.baseLearnerOption));
			classifier.resetLearning();

			if (this.ensemble.size() > 0) {
				double et = 0;
				// Reading all data chunk instances
				for (int i = 0; i < mt; i++) {
					// Compute error of the existing ensemble on new data
					boolean vote = this.correctlyClassifies(this.buffer
							.instance(i));
					if (!vote) {
						et += 1.0 / mt;
					}
				}
				// Normalizing error
				double weightSum = 0.0;
				// Reading all data chunk instances
				for (int i = 0; i < mt; i++) {
					Instance instance = this.buffer.instance(i);
					// Updating instance weights
					boolean vote = this.correctlyClassifies(instance);
					double error = (1.0 / mt) * (vote ? et : 1.0);
					instance.setWeight(error);
					weightSum += error;
				}
				// Reading all data chunk instances
				for (int i = 0; i < mt; i++) {
					Instance instance = this.buffer.instance(i);
					// Normalize weights
					instance.setWeight(instance.weight() / weightSum);
					
					// Call base classifier
					Instance trainingInstance = (Instance)instance.copy();
					trainingInstance.setWeight(1);
					classifier.trainOnInstance(trainingInstance);
				}
			} else {
				// First run! Iterating through all instances in the data chunk
				for (int i = 0; i < mt; i++) {
					Instance instance = this.buffer.instance(i);

					// Initialize weights
					instance.setWeight(1.0 / mt);

					// Call base classifier
					Instance trainingInstance = (Instance)instance.copy();
					trainingInstance.setWeight(1);
					classifier.trainOnInstance(trainingInstance);
				}
			}
			this.ensemble.add(classifier.copy());
			this.bkts.add(new ArrayList<Double>());
			this.wkts.add(new ArrayList<Double>());
			this.ensembleWeights.clear();
			int t = this.ensemble.size();
			this.maxError = Double.MIN_VALUE;
			this.errorIndex = 1;
			// Evaluate all existing classifiers on new data set
			for (int k = 1; k <= t; k++) {
				double ekt = 0;
				// Reading all data chunk instances
				for (int i = 0; i < mt; i++) {
					Instance instance = this.buffer.instance(i);
					if (!this.ensemble.get(k - 1).correctlyClassifies(instance)) {
						// Ensemble incorrectly classifies this instance
						ekt += instance.weight();
					}
				}
				if (k == t && ekt > 0.5) {
					// Generate a new classifier
					Classifier c = (Classifier) getPreparedClassOption(this.baseLearnerOption);
					c.resetLearning();
					this.ensemble.set(k - 1, c);
				} else if (ekt > 0.5) {
					// Remove voting power of this classifier
					ekt = 0.5;
				}
				// Storing the index of the classifier with higher error in case
				// of error-based pruning
				if (ekt > maxError) {
					maxError = ekt;
					errorIndex = k;
				}
				// Normalizing errors
				double bkt = ekt / (1.0 - ekt);
				// Retrieving normalized errors for this classifier
				ArrayList<Double> nbkt = this.bkts.get(k - 1);
				nbkt.add(bkt);
				// Compute the weighted average of all normalized errors for kth
				// classifier h_k
				double wkt = 1.0 / (1.0 + Math.exp(-this.slope
						* (t - k - this.crossingPoint)));
				List<Double> weights = this.wkts.get(k - 1);
				double sum = 0;
				for (int j = 0; j < weights.size(); j++) {
					sum += weights.get(j);
				}
				weights.add(wkt / (sum + wkt));
				double sbkt = 0.0;
				for (int j = 0; j < weights.size(); j++) {
					sbkt += weights.get(j) * nbkt.get(j);
				}
				// Calculate classifier voting weights
				this.ensembleWeights.add(Math.log(1.0 / sbkt));
			}
			// Ensemble pruning strategy				
			if (pruning == 1 && t > ensembleSize) { // Age-based
				this.ensemble.remove(0);
				this.ensembleWeights.remove(0);
				this.bkts.remove(0);
				this.wkts.remove(0);
			} else if (pruning == 2 && t > ensembleSize) { // Error-based
				this.ensemble.remove(errorIndex - 1);
				this.ensembleWeights.remove(errorIndex - 1);
				this.bkts.remove(errorIndex - 1);
				this.wkts.remove(errorIndex - 1);
			}
			this.buffer = new Instances(this.getModelContext());
		}
	}

	@Override
	public boolean isRandomizable() {
		return false;
	}

	@Override
	public double[] getVotesForInstance(Instance inst) {
		DoubleVector combinedVote = new DoubleVector();
		if (this.trainingWeightSeenByModel > 0.0) {
			for (int i = 0; i < this.ensemble.size(); i++) {
				if (this.ensembleWeights.get(i) > 0.0) {
					DoubleVector vote = new DoubleVector(this.ensemble.get(i)
							.getVotesForInstance(inst));
					if (vote.sumOfValues() > 0.0) {
						vote.normalize();
						vote.scaleValues(this.ensembleWeights.get(i));
						combinedVote.addValues(vote);
					}
				}
			}
		}
		return combinedVote.getArrayRef();
	}

	@Override
	public void getModelDescription(StringBuilder out, int indent) {
	}

	@Override
	protected Measurement[] getModelMeasurementsImpl() {
		return null;
	}
}
