/*
 *    DriftWeightedClassificationPerformanceEvaluator.java
 *    Copyright (C) 2007 University of Waikato, Hamilton, New Zealand
 *    @author Dariusz Brzezinski (dariusz.brzezinski at cs.put.poznan.pl)
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program. If not, see <http://www.gnu.org/licenses/>.
 *    
 */
package moa.evaluation;

import moa.classifiers.Classifier;
import moa.classifiers.core.driftdetection.DriftDetectionMethod;
import moa.classifiers.drift.SingleClassifierDrift;
import moa.core.Measurement;
import moa.core.ObjectRepository;
import moa.options.ClassOption;
import moa.options.FloatOption;
import moa.options.AbstractOptionHandler;
import moa.tasks.TaskMonitor;

import weka.core.Instance;
import weka.core.Utils;

/**
 * Classification evaluator that weights evaluation results according to their distance from the last drift.
 *
 * @author Dariusz Brzezinski (dariusz.brzezinski at cs.put.poznan.pl)
 */
public class DriftWeightedClassificationPerformanceEvaluator extends AbstractOptionHandler
        implements ClassificationPerformanceEvaluator {

    private static final long serialVersionUID = 1L;

    protected double TotalweightObserved;

    public FloatOption averageDriftWeightOption = new FloatOption("avgDriftWeight",
            'w', "Average weight for examples during a drift", 3.0);

    public ClassOption baseLearnerOption = new ClassOption("baseLearner", 'l',
            "Classifier to train.", Classifier.class, "bayes.NaiveBayes");
    
    public ClassOption driftDetectionMethodOption = new ClassOption("driftDetectionMethod", 'd',
             "Drift detection method to use.", DriftDetectionMethod.class, "DDM");
    
    protected Estimator weightCorrect;

    protected SingleClassifierDrift driftDetector;
    
    protected class Estimator {

        protected double weight;
        
        protected double averageDriftWeight;

        protected double totalWeight;
        
        protected double weightedEstimation;
        
        protected int timeSinceLastDrift;
        
        protected int detectedDrifts;
        
        protected double logBase;
        
        protected int averageTimeBetweenDrifts;
        
        protected int numClassified;
        
        protected int numCorrectlyClassified;

        public Estimator(double avgDriftWeight) {
        	weight = 1;
        	averageDriftWeight = avgDriftWeight;
            weightedEstimation = 0;
            totalWeight = 0;
            timeSinceLastDrift = 0;
            detectedDrifts = 0;
            logBase = Math.log(Math.pow(Math.E, 1/(averageDriftWeight-1)));
            numClassified = 0;
            numCorrectlyClassified =0;
        }

        public void add(double value, boolean driftDetected) {
        	if (driftDetected){
        		if (detectedDrifts == 0) {
        			averageTimeBetweenDrifts = timeSinceLastDrift;
        		} else {
        			averageTimeBetweenDrifts = (averageTimeBetweenDrifts * detectedDrifts + timeSinceLastDrift) / (detectedDrifts + 1);
        		}
        		timeSinceLastDrift = 1;
        		detectedDrifts++;
        	} else {
        		timeSinceLastDrift++;
        	}
        	
        	weight = calculateWeight();
        	totalWeight += weight;
            weightedEstimation += weight * value;
            numClassified++;
            numCorrectlyClassified += value;
        }

        public double weightedEstimation() {
            return totalWeight > 0 ? weightedEstimation / totalWeight : 0.0;
        }
        
        public double estimation() {
            return numClassified > 0 ? (1.0 * numCorrectlyClassified) / (1.0 * numClassified) : 0.0;
        }
        
        private double calculateWeight() {
        	if (detectedDrifts > 0){
        		return Math.max(-(Math.log(timeSinceLastDrift)/logBase) + 1 + Math.log(averageTimeBetweenDrifts)/logBase, 1);
        	} else {
        		return 1.0;
        	}
        }
    }
    
    @Override
    public void reset() {
    	driftDetector = new SingleClassifierDrift();
    	driftDetector.baseLearnerOption.setValueViaCLIString(this.baseLearnerOption.getValueAsCLIString());
    	driftDetector.driftDetectionMethodOption.setValueViaCLIString(this.driftDetectionMethodOption.getValueAsCLIString());
    	driftDetector.prepareForUse();
    	driftDetector.resetLearning();
        weightCorrect = new Estimator(this.averageDriftWeightOption.getValue());
    }

    @Override
    public void addResult(Instance inst, double[] classVotes) {
        double weight = inst.weight();
        int trueClass = (int) inst.classValue();
        if (weight > 0.0) {
            this.TotalweightObserved += weight;
            driftDetector.trainOnInstance(inst);
            if (Utils.maxIndex(classVotes) == trueClass) {
                this.weightCorrect.add(1, driftDetector.isChangeDetected());
            } else {
                this.weightCorrect.add(0, driftDetector.isChangeDetected());
            }
        }
    }

    @Override
    public Measurement[] getPerformanceMeasurements() {
        return new Measurement[]{
                    new Measurement("classified instances",
                    this.TotalweightObserved),
                    new Measurement("accuracy (percent)",
                    getFractionCorrectlyClassified() * 100.0),
                    new Measurement("weighted accuracy (percent)",
                    getWeightedFractionCorrectlyClassified() * 100.0)};
    }

    public double getTotalWeightObserved() {
        return this.TotalweightObserved;
    }

    public double getFractionCorrectlyClassified() {
        return this.weightCorrect.estimation();
    }

    public double getFractionIncorrectlyClassified() {
        return 1.0 - getFractionCorrectlyClassified();
    }
    
    public double getWeightedFractionCorrectlyClassified() {
        return this.weightCorrect.weightedEstimation();
    }

    public double getWeightedFractionIncorrectlyClassified() {
        return 1.0 - getWeightedFractionCorrectlyClassified();
    }

    @Override
    public void getDescription(StringBuilder sb, int indent) {
        Measurement.getMeasurementsDescription(getPerformanceMeasurements(),
                sb, indent);
    }

    @Override
    public void prepareForUseImpl(TaskMonitor monitor,
            ObjectRepository repository) {
        reset();
    }
}
